/// @file
/// Test creation, action, and destruction for mass matrix operator
/// \test Test creation, action, and destruction for mass matrix operator
#include <ceed.h>
#include <stdlib.h>
#include <math.h>

#include "t500-operator.h"

int main(int argc, char **argv) {
  Ceed ceed;
  CeedElemRestriction elem_restr_x, elem_restr_u, elem_restr_qd_i;
  CeedBasis basis_x, basis_u;
  CeedQFunction qf_setup, qf_mass;
  CeedOperator op_setup, op_mass;
  CeedVector q_data, X, U, V;
  const CeedScalar *hv;
  CeedInt num_elem = 15, P = 5, Q = 8;
  CeedInt num_nodes_x = num_elem+1, num_nodes_u = num_elem*(P-1)+1;
  CeedInt ind_x[num_elem*2], ind_u[num_elem*P];
  CeedScalar x[num_nodes_x];

//! [Ceed Init]
  CeedInit(argv[1], &ceed);
//! [Ceed Init]
  for (CeedInt i=0; i<num_nodes_x; i++)
    x[i] = (CeedScalar) i / (num_nodes_x - 1);
  for (CeedInt i=0; i<num_elem; i++) {
    ind_x[2*i+0] = i;
    ind_x[2*i+1] = i+1;
  }
//! [ElemRestr Create]
  CeedElemRestrictionCreate(ceed, num_elem, 2, 1, 1, num_nodes_x, CEED_MEM_HOST,
                            CEED_USE_POINTER, ind_x, &elem_restr_x);
//! [ElemRestr Create]

  for (CeedInt i=0; i<num_elem; i++) {
    for (CeedInt j=0; j<P; j++) {
      ind_u[P*i+j] = i*(P-1) + j;
    }
  }
//! [ElemRestrU Create]
  CeedElemRestrictionCreate(ceed, num_elem, P, 1, 1, num_nodes_u, CEED_MEM_HOST,
                            CEED_USE_POINTER, ind_u, &elem_restr_u);
  CeedInt strides_qd[3] = {1, Q, Q};
  CeedElemRestrictionCreateStrided(ceed, num_elem, Q, 1, Q*num_elem, strides_qd,
                                   &elem_restr_qd_i);
//! [ElemRestrU Create]

//! [Basis Create]
  CeedBasisCreateTensorH1Lagrange(ceed, 1, 1, 2, Q, CEED_GAUSS, &basis_x);
  CeedBasisCreateTensorH1Lagrange(ceed, 1, 1, P, Q, CEED_GAUSS, &basis_u);
//! [Basis Create]

//! [QFunction Create]
  CeedQFunctionCreateInterior(ceed, 1, setup, setup_loc, &qf_setup);
  CeedQFunctionAddInput(qf_setup, "weight", 1, CEED_EVAL_WEIGHT);
  CeedQFunctionAddInput(qf_setup, "dx", 1, CEED_EVAL_GRAD);
  CeedQFunctionAddOutput(qf_setup, "rho", 1, CEED_EVAL_NONE);

  CeedQFunctionCreateInterior(ceed, 1, mass, mass_loc, &qf_mass);
  CeedQFunctionAddInput(qf_mass, "rho", 1, CEED_EVAL_NONE);
  CeedQFunctionAddInput(qf_mass, "u", 1, CEED_EVAL_INTERP);
  CeedQFunctionAddOutput(qf_mass, "v", 1, CEED_EVAL_INTERP);
//! [QFunction Create]

//! [Setup Create]
  CeedOperatorCreate(ceed, qf_setup, CEED_QFUNCTION_NONE, CEED_QFUNCTION_NONE,
                     &op_setup);
//! [Setup Create]

//! [Operator Create]
  CeedOperatorCreate(ceed, qf_mass, CEED_QFUNCTION_NONE, CEED_QFUNCTION_NONE,
                     &op_mass);
//! [Operator Create]

  CeedVectorCreate(ceed, num_nodes_x, &X);
  CeedVectorSetArray(X, CEED_MEM_HOST, CEED_USE_POINTER, x);
  CeedVectorCreate(ceed, num_elem*Q, &q_data);

//! [Setup Set]
  CeedOperatorSetField(op_setup, "weight", CEED_ELEMRESTRICTION_NONE, basis_x,
                       CEED_VECTOR_NONE);
  CeedOperatorSetField(op_setup, "dx", elem_restr_x, basis_x, CEED_VECTOR_ACTIVE);
  CeedOperatorSetField(op_setup, "rho", elem_restr_qd_i, CEED_BASIS_COLLOCATED,
                       CEED_VECTOR_ACTIVE);
//! [Setup Set]

//! [Operator Set]
  CeedOperatorSetField(op_mass, "rho", elem_restr_qd_i,CEED_BASIS_COLLOCATED,
                       q_data);
  CeedOperatorSetField(op_mass, "u", elem_restr_u, basis_u, CEED_VECTOR_ACTIVE);
  CeedOperatorSetField(op_mass, "v", elem_restr_u, basis_u, CEED_VECTOR_ACTIVE);
//! [Operator Set]

//! [Setup Apply]
  CeedOperatorApply(op_setup, X, q_data, CEED_REQUEST_IMMEDIATE);
//! [Setup Apply]

  CeedVectorCreate(ceed, num_nodes_u, &U);
  CeedVectorSetValue(U, 0.0);
  CeedVectorCreate(ceed, num_nodes_u, &V);
//! [Operator Apply]
  CeedOperatorApply(op_mass, U, V, CEED_REQUEST_IMMEDIATE);
//! [Operator Apply]

  CeedVectorGetArrayRead(V, CEED_MEM_HOST, &hv);
  for (CeedInt i=0; i<num_nodes_u; i++)
    if (fabs(hv[i]) > 1e-14) printf("[%d] v %g != 0.0\n",i, hv[i]);
  CeedVectorRestoreArrayRead(V, &hv);

  CeedQFunctionDestroy(&qf_setup);
  CeedQFunctionDestroy(&qf_mass);
  CeedOperatorDestroy(&op_setup);
  CeedOperatorDestroy(&op_mass);
  CeedElemRestrictionDestroy(&elem_restr_u);
  CeedElemRestrictionDestroy(&elem_restr_x);
  CeedElemRestrictionDestroy(&elem_restr_qd_i);
  CeedBasisDestroy(&basis_u);
  CeedBasisDestroy(&basis_x);
  CeedVectorDestroy(&X);
  CeedVectorDestroy(&U);
  CeedVectorDestroy(&V);
  CeedVectorDestroy(&q_data);
  CeedDestroy(&ceed);
  return 0;
}
